/*
 * (c) 2006 Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#define _POSIX_C_SOURCE 2

/* We need SUS/XPG4v2 for readlink.
 * This is actually not a requirement, as it is only used for debug
 * purposes. If you don't have this, get comment out resolve_symlink and
 * define it as a function always return 0.
 */
#define _XOPEN_SOURCE 500

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "log.h"
#include "tokens.h"


typedef struct arg_s {
	char *arg;
	struct arg_s *prev;
} arg_t;

typedef struct job_s {
	pid_t pid;
	char **args;
	size_t numargs;
	char *in, *out, *err;
	int outfile;
	int append;	/* 1 = stdout, 2 = stderr, 3 = both */
	struct job_s *next;
} job_t;

static arg_t *Baseargs = 0;
static size_t Numbaseargs = 0, Running = 0, Waiting = 0;
static unsigned long Limit = 0;
static job_t *Executing = 0, *Ready = 0, *ReadyEnd = 0;
static int Stdout = -1, Stderr = -1, Stdin = -1, Prompt = 0;
static FILE *Tty = 0;
static char *Path;

extern int yylex(void);
extern verbose_t Verbose;


static void *Malloc(size_t s)
{
	int tries = 0, time = 5;
	for (;;) {
		void *r = malloc(s);
		if (r != 0)
			return r;
		warn("out of memory - retrying in %d seconds\n",time);
		(void) sleep(time);
		time += 5;
		++tries;
		if (tries == 10) {
			warn("Persisiting low memory situation: stopping.\n"
				"Please send SIGCONT to contiue.\n");
			if (-1 == raise(SIGSTOP))
				abort();
			time = 5;
			tries = 0;
		}
	}
}


static char *Strdup(const char *s)
{
	size_t l = strlen(s);
	return memcpy(Malloc(l+1),s,l+1);
}


static void exec_job(job_t *j)
{
	pid_t pid;
	int in = -1, out = -1, err = -1;
	if (Prompt) {
		char y[16];
		char **a = j->args;
		fprintf(Tty,"start job %s", *a++);
		while(*a)
			fprintf(Tty," %s",*a++);
		fprintf(Tty,"?\n");
		fgets(y,15,Tty);
		if (ferror(Tty))
			error("cannot read from terminal\n");
		if (strcasecmp(y,"yes\n") && strcasecmp(y,"y\n")) {
			j->pid = 0;
			return;
		}
	}
	if (j->in) {	/* stdin is redirected */
		in = open(j->in,O_WRONLY|O_CREAT|O_TRUNC,0666);
		if (in == -1) {
			warn("cannot open input file %s: %s\n",j->in,strerror(errno));
			return;
		} else
			dbug("job input redirected to %s\n",j->in);
	} else {
		in = Stdin;
	}
	if (j->err) {	/* stdout and stderr are redirected */
		int mode;
	       	if (j->append & 2)
			mode = O_WRONLY|O_APPEND;
		else
			mode = O_WRONLY|O_CREAT|O_TRUNC|O_EXCL;
		err = open(j->err,mode,0666);
		if (err == -1) {
			warn("cannot open error output file %s: %s\n",j->err,strerror(errno));
			return;
		} else {
			out = err;
			dbug("job error output redirected to %s\n",j->err);
		}
	} else if (Stdout != -1) {
		err = Stderr;
	}
	if (j->out) {	/* output is redirected */
		int mode;
	       	if (j->append & 1)
			mode = O_WRONLY|O_APPEND;
		else
			mode = O_WRONLY|O_CREAT|O_TRUNC|O_EXCL;
		out = open(j->out,mode,0666);
		if (out == -1) {
			warn("cannot open output file %s: %s\n",j->out,strerror(errno));
			return;
		} else
			dbug("job output redirected to %s\n",j->out);
	} else if (Stdout != -1) {
		out = Stdout;
	}
	if (err == -1) {
		char fname[] = "/tmp/xjobs-XXXXXX";
		err = mkstemp(fname);
		if (-1 == err)
			error("cannot create intermediate output file %s: %s\n",fname,strerror(errno));
		j->outfile = open(fname,O_RDONLY);
		if (-1 == j->outfile)
			error("cannot access intermediate output file: %s\n",strerror(errno));
		(void) unlink(fname);
		if (out == -1)
			out = err;
	}
	pid = fork();
	if (pid == -1)
		error("error could not fork: %s\n",strerror(errno));
	if (pid != 0) {
		char **a = j->args;
		j->pid = pid;
		assert(a);
		info("### started pid %d: %s",pid,*a++);
		while(*a)
			info(" %s",*a++);
		info("\n");
		j->next = Executing;
		if ((in != -1) && (in != Stdin))
			(void) close(in);
		if ((out != -1) && (out != Stdout))
			(void) close(out);
		if ((err != -1) && (err != Stderr))
			(void) close(err);
		Executing = j;
		if (--Waiting)
			dbug("%d jobs still waiting in the queue\n",Waiting);
		++Running;
	} else {
		int r, i;
		struct rlimit rl;
		if (j->outfile != -1)
			(void) close(j->outfile);
		if (in != -1) {
			(void) close(STDIN_FILENO);
			if (-1 == dup2(in,STDIN_FILENO))
				warn("unable to redirect stdin: %s\n",strerror(errno));
		}
		if (out != -1) {
			(void) close(STDOUT_FILENO);
			if (-1 == dup2(out,STDOUT_FILENO))
				warn("unable to redirect stdout: %s\n",strerror(errno));
		}
		if (err != -1) {
			(void) close(STDERR_FILENO);
			if (-1 == dup2(err,STDERR_FILENO))
				warn("unable to redirect stderr: %s\n",strerror(errno));
		}
		r = getrlimit(RLIMIT_NOFILE, &rl);
		assert(r == 0);
		for (i = 3; i < rl.rlim_max; i++)
			(void) close(i);
		r = execv(j->args[0],j->args);
		assert(r == -1);
		error("failed to execute %s: %s\n",j->args[0],strerror(errno));
		exit(EXIT_FAILURE);
	}
}


static int resolve_symlink(char *cmd)
{
	char lt[PATH_MAX+1];
	ssize_t l;
	/* handle symlinks */
	bzero(lt,sizeof(lt));
	while (-1 != (l = readlink(cmd,lt,PATH_MAX))) {
		lt[l] = '\0';
		dbug("is symlink: %s\n",lt);
		/* handle relative links */
		if (lt[0] != '/') {
			char *base = strrchr(cmd,'/');   
			if (base == 0) {
				base = cmd;
			} else if ((strlen(lt) + (base - cmd)) > PATH_MAX) {
				warn("command exceed PATH_MAX limit of your system\n");
				return -1;
			}
			++base;
			(void) strcpy(base,lt);
			dbug("resolved to: %s\n",cmd);
		} else
			(void) strcpy(cmd,lt);
	}
	return 0;
}


static char *complete_exe(char *exe)
{
	char *token, *next;
	size_t exel;

	if ((exe[0] == '/') && (0 == access(exe,X_OK)))
                return exe;
	exel = strlen(exe);
	if ((strncmp(exe,"./",2) == 0) || (strncmp(exe,"../",3) == 0)) {
		if (0 == access(exe,X_OK))
			return exe;
		if (Verbose == Debug) {
			char cmd[PATH_MAX+1];
			(void) strcpy(cmd,exe);
			(void) resolve_symlink(cmd);
			return 0;
		}
	}
	token = Path;
	next = strchr(Path,':');
        while (token && *token) {
		char cmd[PATH_MAX+1];
		size_t off;

		if (next)
			*next = 0;
		if ((off = strlen(token)) > PATH_MAX) {
			warn("command exceed PATH_MAX limit of your system\n");
			return 0;
		}
		(void) strcpy(cmd, token);
		if (cmd[off-1] != '/') {
			cmd[off] = '/';
			++off;
		}
		if (exel + off > PATH_MAX) {
			warn("command exceed PATH_MAX limit of your system\n");
			return 0;
		}
		(void) strcpy(cmd+off, exe);
		if (next) {
			*next = ':';
			token = next + 1;
			next = strchr(token,':');
		} else
			token = 0;
		if (-1 == access(cmd,X_OK))
			continue;
		dbug("candidate: %s\n",cmd);
		if ((Verbose == Debug) && resolve_symlink(cmd))
			return 0;
		if (0 == access(cmd,X_OK)) {
			dbug("using: %s\n",cmd);
			free(exe);
                        return Strdup(cmd);
		}
        }
        return 0;
}


static void start_jobs(void)
{
	while ((Running < Limit) && Ready) {
		job_t *j = Ready;
		if ((j->args == 0) && (Running > 0)) {
			/* sequence point */
			dbug("sequence point - waiting for all processes to finish\n");
			return;
		}
		Ready = j->next;
		Ready = j->next;
		if (Ready == 0)
			ReadyEnd = 0;
		if (j->args) {
			exec_job(j);
			if (j->pid == 0) {
				char **a = j->args;
				info("### skipping job: %s",a);
				assert(a);
				while(*a)
					warn(" %s",*a++);
				warn("\n");
			}
			free(j->out);
			free(j->err);
			free(j->in);
		} else {
			dbug("starting jobs after sequence point\n");
			free(j);
		}
	}
}


static void add_job(arg_t *a, size_t n, char *in, char *out, char *err, int app)
{
	char **argv, *exe;
	int i = n + Numbaseargs + 1;
	job_t *j;

	assert(a);
	assert(n > 0);
	j = Malloc(sizeof(job_t));
	bzero(j,sizeof(job_t));
	argv = Malloc(sizeof(char *) * i);
	j->in = in;
	j->err = err;
	j->out = out;
	j->append = app;
	j->outfile = -1;
	j->args = argv;
	--i;
	argv[i] = 0;
	j->numargs = i;
	while (n) {
		arg_t *c = a;
		assert(a);
		a = c->prev;
		--n;
		--i;
		argv[i] = c->arg;
		free(c);
	}
	assert(a == 0);
	a = Baseargs;
	while (i) {
		arg_t *c = a;
		assert(a);
		a = c->prev;
		--i;
		argv[i] = Strdup(c->arg);
	}
	assert(a == 0);
	exe = complete_exe(argv[0]);
	if (0 == exe) {
		int i;
		warn("cannot find executable %s\n",argv[0]);
		for (i = 0; i < j->numargs; ++i)
			free(argv[i]);
		free(argv);
		free(j);
		return;
	}
	argv[0] = exe;
	while (*argv) {
		dbug("argument %d: <%s>\n",argv - j->args, *argv);
		++argv;
	}
	j->next = 0;
	if (ReadyEnd)
		ReadyEnd->next = j;
	else
		Ready = j;
	ReadyEnd = j;
	++Waiting;
	start_jobs();
	if (Ready)
		dbug("job queued\n");
}


void clear_job(pid_t pid, int ret)
{
	int i;
	job_t *j = Executing, *lj = 0;

	while (j && (j->pid != pid)) {
		lj = j;
		j = j->next;
	}
	assert(j);
	if (lj)
		lj->next = j->next;
	else
		Executing = j->next;
	if (j->outfile != -1) {
		char **a = j->args;
		(void) printf("### output of pid %ld: %s",pid,*a);
		++a;
		while (*a) {
			(void) printf(" %s",*a);
			++a;
		}
		(void) printf("\n");
		struct stat st;
		int ret;
		void *mem;
		ret = fstat(j->outfile,&st);
		assert(ret == 0);
		if (st.st_size > 0) {
			mem = mmap(0,st.st_size,PROT_READ,MAP_SHARED,j->outfile,0);
			if (mem != (void *)-1) {
				int w = write(STDOUT_FILENO,mem,st.st_size);
				assert(w == st.st_size);
			} else
				warn("could not mmap output of child: %s\n",strerror(errno));
			(void) munmap(mem,st.st_size);
		}
		(void) close(j->outfile);
		(void) printf("### exitcode %d\n",ret);
	} else {
		char **a = j->args;
		info("### pid %6d, exitcode %d (%s",pid,ret,*a);
		++a;
		while (*a) {
			info(" %s",*a);
			++a;
		}
		info(")\n",ret);
	}
	for (i = 0; i < j->numargs; ++i)
		free(j->args[i]);
	free(j->args);
	free(j);
	--Running;
}


static void usage(void)
{
	(void) printf(
	"xjobs [options] [command [argument ...]]\n"
	"========================================\n"
	"valid options are:\n"
	"-h           : print this usage information and exit\n"
	"-j <num>     : maximum number of jobs to execute in parallel\n"
	"-s <file>    : script to execute (default: read from stdin)\n"
	"-l <num>     : combine <num> lines to a single job\n"
	"-n           : redirect stdout and stderr of childs to /dev/null\n"
	"-v <level>   : set verbosity to level\n"
	"               (0=silent,1=error,2=warning,3=info,4=debug)\n"
	"-V           : print version and exit\n"
	);
	exit(0);
}


static void version(void)
{
	(void) printf(
	"xjobs version 20060411\n"
	"Copyright 2006, Thomas Maier-Komor\n"
	"License: GPLv2\n"
	);
	exit(0);
}


int main(int argc, char **argv)
{
	typedef enum { init, std, white, dquoted, squoted, desc, sesc } state_t;
	int i, append = 0;
       	unsigned long lines = 1, lcount = 0;
	char *jobin = 0, *jobout = 0, *joberr = 0;
	extern char *optarg;
	extern int optind, optopt;
	arg_t *args = 0;
	size_t numargs = 0;
	token_t t;

	Path = getenv("PATH");
	while ((i = getopt(argc,argv,"hj:l:nps:v:V")) != -1) {
		switch (i) {
		default:
			abort();
			break;
		case 'h':
			usage();
			break;
		case 'j':
			if (1 == sscanf(optarg,"%lu",&Limit) && (Limit > 0))
				dbug("maximum number of jobs set to %d\n",Limit);
			else
				error("error in argument to option -j\n");
			break;
		case 'l':
			if (1 == sscanf(optarg,"%lu",&lines) || (lines > 0))
				dbug("combining %lu lines to a single command\n");
			else
				error("error in argument to option -l\n");
			break;
		case 'n':
			Stdout = open("/dev/null",O_WRONLY);
			if (Stdout == -1)
				error("could not open /dev/null: %s\n",strerror(errno));
			else
				dbug("output redirected to /dev/null\n");
			Stderr = Stdout;
			break;
		case 'p':
			Prompt = 1;
			Tty = fopen("/dev/tty","r+");
			if (Tty == 0)
				error("Cannot open terminal for prompt mode: %s\n",strerror(errno));
			dbug("enabling prompt mode\n");
			break;
		case 's':
			{
				extern FILE *yyin;
				FILE *in = fopen(optarg,"r");
				if (in == 0)
					error("could not open input file %s: %s\n",optarg,strerror(errno));
				yyin = in;
				dbug("input file set to %s\n",optarg);
				break;
			}
		case 'v':
			if (1 == sscanf(optarg,"%ld",&Verbose)) {
				if (Verbose < Silent || Verbose > Debug) {
					error("value out of range for option -v\n");
					Verbose = 0;
				}
				dbug("verbose set to %d\n",Verbose);
			} else
				Verbose = 0;
			break;
		case 'V':
			version();
			break;
		case '?':
			error("unkonwn option -%c\n",optopt);
			break;
		case ':':
			error("option -%c requires an operand\n",optopt);
			break;
		}
	}
        dbug("PATH=\"%s\"\n", Path);
	if (Limit == 0) {
		Limit = sysconf(_SC_NPROCESSORS_ONLN);
		dbug("%d processors currently online (default job limit)\n",Limit);
	}
	if (Stdin == -1) {	/* no default input redirector */
		Stdin = open("/dev/null",O_RDONLY);
		if (Stdin == -1)
			error("could not open /dev/null: %s\n",strerror(errno));
	}

	while (optind < argc) {
		arg_t *a = Malloc(sizeof(arg_t));
		a->arg = argv[optind];
		if (0 == strcmp("<",argv[optind])) {
			Stdin = open(argv[++optind],O_RDONLY);
			if (-1 == Stdin)
				error("could not open default stdin for jobs (%s): %s\n",argv[optind],strerror(errno));
		} else if (0 == strcmp(">",argv[optind])) {
			Stdout = open(argv[++optind],O_WRONLY|O_CREAT|O_EXCL,0666);
			if (-1 == Stdout)
				error("could not open stdout for jobs (%s): %s\n",argv[optind],strerror(errno));
		} else if (0 == strcmp(">>",argv[optind])) {
			Stdout = open(argv[++optind],O_WRONLY|O_APPEND);
			if (-1 == Stdout)
				error("could not append to stdout for jobs (%s): %s\n",argv[optind],strerror(errno));
		} else if (0 == strcmp(">&",argv[optind])) {
			Stderr = Stdout = open(argv[++optind],O_WRONLY|O_CREAT|O_EXCL,0666);
			if (-1 == Stdout)
				error("could not open stdout/stderr for jobs (%s): %s\n",argv[optind],strerror(errno));
		} else if (0 == strcmp(">>&",argv[optind])) {
			Stdout = open(argv[++optind],O_WRONLY|O_APPEND);
			if (-1 == Stdout)
				error("could not append to stdout/stderr for jobs (%s): %s\n",argv[optind],strerror(errno));
		} else {
			a->prev = Baseargs;
			Baseargs = a;
			++Numbaseargs;
		}
		++optind;
	}

	while (t = yylex()) {
		extern const char *Buffer;
		pid_t pid;
		int ret;

		switch (t) {
		default:
			abort();
			break;
		case SEQUPT:
			{
				job_t *j = malloc(sizeof(job_t));
				j->args = 0;
				if (ReadyEnd)
					ReadyEnd->next = j;
				else
					Ready = j;
				ReadyEnd = j;
				break;
			}
		case INPUT:
			t = yylex();
			if ((t != UQUOTED) && (t != DQUOTED) && (t != SQUOTED)) {
				warn("invalid argument for input redirector <");
				do t = yylex(); while (t && (t != EOL));
			} else {
				jobin = Strdup(Buffer);
			}
			break;
		case APPEND:
			append |= 1;
			/*FALLTHROUGH*/
		case OUTPUT:
			t = yylex();
			if ((t != UQUOTED) && (t != DQUOTED) && (t != SQUOTED)) {
				warn("invalid argument for output redirector >");
				do t = yylex(); while (t && (t != EOL));
			} else {
				jobout = Strdup(Buffer);
			}
			break;
		case APPERR:
			append = 3;
			/*FALLTHROUGH*/
		case ERRLOG:
			t = yylex();
			if ((t != UQUOTED) && (t != DQUOTED) && (t != SQUOTED)) {
				warn("invalid argument for log redirector >&");
				do t = yylex(); while (t && (t != EOL));
			} else {
				joberr = Strdup(Buffer);
			}
			break;
		case UQUOTED:
		case DQUOTED:
		case SQUOTED:
			{
				arg_t *a = Malloc(sizeof(arg_t));
				bzero(a,sizeof(arg_t));
				a->arg = Strdup(Buffer);
				a->prev = args;
				args = a;
				++numargs;
				break;
			}
		case EOL:
			if (++lcount < lines)
				break;
			if (numargs)
				add_job(args,numargs,jobin,jobout,joberr,append);
			args = 0;
			numargs = 0;
			jobin = 0;
			jobout = 0;
			joberr = 0;
			append = 0;
			lcount = 0;
			break;
		case ERR:
			info("ignoring erroneous line\n");
			break;
		}
		
		if (Executing && (pid = waitpid(0,&ret,WNOHANG))) {
			if (pid != -1) {
				clear_job(pid,ret);
			}
		}
		start_jobs();
	}
	if (numargs)
		add_job(args,numargs,jobin,jobout,joberr,append);

	while (Executing || Ready) {
		pid_t pid;
		int ret;
		start_jobs();
		pid = waitpid(0,&ret,0);
		clear_job(pid,ret);
	}
	return 0;
}

